@@ -6,9 +6,10 @@ from appium import webdriver
 from psutil import Process, NoSuchProcess
 from robot.api.deco import keyword
 from robot.libraries.BuiltIn import BuiltIn
-from selenium.common.exceptions import NoSuchElementException
+from selenium.common.exceptions import NoSuchElementException, InvalidSelectorException
 from selenium.webdriver.common.action_chains import ActionChains
-from time import sleep
+from time import sleep, time
+from robot import utils
 
 try:
     AppiumCommon = importlib.import_module('Helpers.AppiumCommon', package='Helpers')
@@ -70,6 +71,9 @@ class DesktopLibrary(AppiumLibrary):
     |       class        |              ClassName             | Click Element `|` class=UIAPickerWheel            |
     |       name         |                Name                | Click Element `|` name=my_element                 |
     |       xpath        |                N/A                 | Click Element `|` xpath=//Button[@Name="Close"]   |
+    |       image        |                N/A                 | Click Element `|` image=file.png                  |
+
+    The ``image`` locator strategy can only be used with Appium v1.18.0 or higher.
 
     Example tests using the windows calculator are located in the tests directory.
 
@@ -110,7 +114,6 @@ class DesktopLibrary(AppiumLibrary):
         This function restricts the keywords used in the library. This is to prevent incompatible
         keywords from imported libraries from being referenced and used.
         """
-        # self.save_appium_screenshot = save_appium_screenshot
         return [
             'maximize_window', 'open_application', 'wait_for_and_clear_text',
             'wait_for_and_click_element', 'wait_for_and_input_password', 'wait_for_and_input_text',
@@ -122,7 +125,8 @@ class DesktopLibrary(AppiumLibrary):
             'mouse_over_and_context_click_element', 'mouse_over_by_offset', 'drag_and_drop',
             'drag_and_drop_by_offset', 'send_keys', 'send_keys_to_element',
             'capture_page_screenshot', 'save_appium_screenshot', 'select_element_from_combobox',
-            'driver_setup', 'driver_teardown',
+            'driver_setup', 'driver_teardown', 'select_elements_from_menu',
+            'select_elements_from_context_menu',
             # External Libraries
             'clear_text', 'click_button', 'click_element', 'close_all_applications',
             'close_application', 'element_attribute_should_match', 'element_should_be_disabled',
@@ -145,12 +149,16 @@ class DesktopLibrary(AppiumLibrary):
         """Starts the WinAppDriver.
 
         ``path`` can be provided if your winappdriver installation is not in the default path of
-        ``C:/Program Files (x86)/Windows Application Driver/WinAppDriver.exe``."""
+        ``C:/Program Files (x86)/Windows Application Driver/WinAppDriver.exe``.
+
+        Not to be used with Appium."""
         self.winappdriver.set_up_driver(path)
 
     @keyword("Driver Teardown")
     def driver_teardown(self):
-        """Stops the WinAppDriver."""
+        """Stops the WinAppDriver.
+
+        Not to be used with Appium."""
         self.winappdriver.tear_down_driver()
 
     @keyword("Maximize Window")
@@ -160,7 +168,8 @@ class DesktopLibrary(AppiumLibrary):
         return True
 
     @keyword("Open Application")
-    def open_application(self, remote_url, alias=None, window_name=None, splash_delay=0, **kwargs):
+    def open_application(self, remote_url, alias=None, window_name=None, splash_delay=0,
+                         exact_match=True, **kwargs):
         """Opens a new application to given Appium server.
         If your application has a splash screen please supply the window name of the final window that will appear.
         For the capabilities of appium server and Windows please check http://appium.io/docs/en/drivers/windows
@@ -170,10 +179,12 @@ class DesktopLibrary(AppiumLibrary):
         | alias               | No     | Alias                                                                |
         | window_name         | No     | Window name you wish to attach, usually after a splash screen        |
         | splash_delay        | No     | Delay used when waiting for a splash screen to load, in seconds      |
+        | exact_match         | No     | Set to False if window_name does not need to match exactly           |
 
         Examples:
         | Open Application | http://localhost:4723/wd/hub | alias=Myapp1         | platformName=Windows            | deviceName=Windows           | app=your.app          |
         | Open Application | http://localhost:4723/wd/hub | alias=Myapp1         | platformName=Windows            | deviceName=Windows           | app=your.app          | window_name=MyApplication          | splash_delay=5          |
+        | Open Application | http://localhost:4723/wd/hub | alias=Myapp1         | platformName=Windows            | deviceName=Windows           | app=your.app          | window_name=MyApplication          | exact_match=False       |
 
         A session for the root desktop will also be opened and can be switched to by running the following:
         | Switch Application | Desktop         |
@@ -181,14 +192,16 @@ class DesktopLibrary(AppiumLibrary):
         desired_caps = kwargs
 
         if window_name:
-            # If the app has a splash screen we need to supply the window_name of the final window. This code path will
-            # start the application and then attach to the correct window via the window_name.
+            # If the app has a splash screen we need to supply the window_name of the final window.
+            # This code path will start the application and then attach to the correct window via
+            # the window_name.
             self._info('Opening application "%s"' % desired_caps['app'])
             subprocess.Popen(desired_caps['app'])
             if splash_delay > 0:
                 self._info('Waiting %s seconds for splash screen' % splash_delay)
                 sleep(splash_delay)
-            return self.switch_application_by_name(remote_url, alias=alias, window_name=window_name, **kwargs)
+            return self.switch_application_by_name(remote_url, alias=alias, window_name=window_name,
+                                                   exact_match=exact_match, **kwargs)
         # global application
         self._open_desktop_session(remote_url)
         application = webdriver.Remote(str(remote_url), desired_caps)
@@ -196,7 +209,8 @@ class DesktopLibrary(AppiumLibrary):
         return self._cache.register(application, alias)
 
     @keyword("Switch Application By Name")
-    def switch_application_by_name(self, remote_url, window_name, alias=None, timeout=5, **kwargs):
+    def switch_application_by_name(self, remote_url, window_name, alias=None, timeout=5,
+                                   exact_match=True, **kwargs):
         """Switches to a currently opened window by ``window_name``.
         For the capabilities of appium server and Windows,
         Please check http://appium.io/docs/en/drivers/windows
@@ -205,24 +219,36 @@ class DesktopLibrary(AppiumLibrary):
         | window_name         | Yes    | Window name you wish to attach        |
         | alias               | No     | alias                                 |
         | timeout             | No     | timeout to connect                    |
+        | exact_match         | No     | Set to False if window_name does not need to match exactly       |
 
         Examples:
         | Switch Application By Name | http://localhost:4723/wd/hub | alias=Myapp1         | platformName=Windows            | deviceName=Windows           | window_name=MyApplication         |
+        | Switch Application By Name | http://localhost:4723/wd/hub | window_name=MyApp    |  exact_match=False  |
 
         A session for the root desktop will also be opened and can be switched to by running the following:
         | Switch Application | Desktop         |
         """
         desired_caps = kwargs
         desktop_session = self._open_desktop_session(remote_url)
+        window_xpath = '//Window[contains(@Name, "' + window_name + '")]'
         try:
-            window = desktop_session.find_element_by_name(window_name)
+            if exact_match:
+                window = desktop_session.find_element_by_name(window_name)
+            else:
+                window = desktop_session.find_element_by_xpath(window_xpath)
             self._debug('Window_name "%s" found.' % window_name)
             window = hex(int(window.get_attribute("NativeWindowHandle")))
         except Exception:
             try:
                 error = "Window '%s' did not appear in <TIMEOUT>" % window_name
-                self._wait_until(timeout, error, desktop_session.find_element_by_name, window_name)
-                window = desktop_session.find_element_by_name(window_name)
+                if exact_match:
+                    self._wait_until(timeout, error, desktop_session.find_element_by_name,
+                                     window_name)
+                    window = desktop_session.find_element_by_name(window_name)
+                else:
+                    self._wait_until(timeout, error, desktop_session.find_element_by_xpath,
+                                     window_xpath)
+                    window = desktop_session.find_element_by_xpath(window_xpath)
                 self._debug('Window_name "%s" found.' % window_name)
                 window = hex(int(window.get_attribute("NativeWindowHandle")))
             except Exception as e:
@@ -283,7 +309,9 @@ class DesktopLibrary(AppiumLibrary):
     def click_element(self, locator):
         """Click element identified by `locator`.
 
-        Supported prefixes: ``accessibility_id``, ``name``, ``class``, ``xpath``
+        Supported prefixes: ``accessibility_id``, ``name``, ``class``, ``xpath``, and ``image``
+
+        The ``image`` locator strategy can only be used with Appium v1.18.0 and newer.
 
         If no prefix is given ``click element`` defaults to ``accessibility_id`` or ``xpath``
         """
@@ -563,7 +591,7 @@ class DesktopLibrary(AppiumLibrary):
         given in absolute format.
 
         `css` can be used to modify how the screenshot is taken. By default
-        the bakground color is changed to avoid possible problems with
+        the background color is changed to avoid possible problems with
         background leaking when the page layout is somehow broken.
 
         See `Save Appium Screenshot` for a screenshot that will be unique across reports
@@ -604,6 +632,64 @@ class DesktopLibrary(AppiumLibrary):
                 self.click_element(element_locator)
             self.switch_application(original_index)
 
+    @keyword("Select Elements From Menu")
+    def select_elements_from_menu(self, *args):
+        """Selects N number of elements in the order they are given. This is useful for working
+        though a nested menu listing of elements.
+
+        On failure this keyword will attempt to select the elements from the desktop session due to
+        the nature of some pop-out menus in Windows."""
+        count = 0
+        try:
+            for each in args:
+                self.click_element(each)
+                count += 1
+        except NoSuchElementException:
+            original_index = self._cache.current_index
+            self.switch_application('Desktop')
+            for each in args[count:]:
+                try:
+                    self.click_element(each)
+                except NoSuchElementException:
+                    self._wait_until_page_contains_element(each, self.get_appium_timeout())
+                    self.click_element(each)
+                count += 1
+            self.switch_application(original_index)
+
+    @keyword("Select Elements From Context Menu")
+    def select_elements_from_context_menu(self, *args):
+        """Context clicks the first element and then selects N number of elements in the order they
+        are given. This is useful for working though a nested context menu listing of elements.
+
+        On failure this keyword will attempt to select the elements from the desktop session due to
+        the nature of some pop-out menus in Windows."""
+        count = 0
+        try:
+            for each in args:
+                if count == 0:
+                    self.mouse_over_and_context_click_element(each)
+                else:
+                    self.click_element(each)
+                count += 1
+        except NoSuchElementException:
+            original_index = self._cache.current_index
+            self.switch_application('Desktop')
+            for each in args[count:]:
+                try:
+                    if count == 0:
+                        self.mouse_over_and_context_click_element(each)
+                    else:
+                        self.click_element(each)
+                except NoSuchElementException:
+                    if count == 0:
+                        self._wait_until_page_contains_element(each, self.get_appium_timeout())
+                        self.mouse_over_and_context_click_element(each)
+                    else:
+                        self._wait_until_page_contains_element(each, self.get_appium_timeout())
+                        self.click_element(each)
+                count += 1
+            self.switch_application(original_index)
+
     # Private
     def _move_to_element(self, actions, element, x_offset=0, y_offset=0):
         if x_offset != 0 or y_offset != 0:
@@ -618,8 +704,8 @@ class DesktopLibrary(AppiumLibrary):
             return self._cache.get_connection(alias)
         except RuntimeError:
             self._debug('Creating new desktop session')
-            desktop_capabilities = dict({"app": "Root", "platformName": "Windows", "deviceName": "Windows",
-                                         "newCommandTimeout": 3600})
+            desktop_capabilities = dict({"app": "Root", "platformName": "Windows",
+                                         "deviceName": "Windows", "newCommandTimeout": 3600})
             desktop_session = webdriver.Remote(str(remote_url), desktop_capabilities)
             self._cache.register(desktop_session, alias=alias)
             return desktop_session
@@ -651,6 +737,36 @@ class DesktopLibrary(AppiumLibrary):
             if first_only:
                 return driver.find_element_by_accessibility_id(criteria)
             return driver.find_elements_by_accessibility_id(criteria)
+        if prefix == 'image':
+            try:
+                if first_only:
+                    return driver.find_element_by_image(criteria)
+                return driver.find_elements_by_image(criteria)
+            except InvalidSelectorException:
+                zoomba.fail("Selecting by image is only available when using Appium "
+                            "v1.18.0 or higher")
+        zoomba.fail("Element locator with prefix '" + prefix + "' is not supported")
+
+    def _is_element_present(self, locator):
+        prefix, criteria = self._parse_locator(locator)
+        driver = self._current_application()
+        if prefix is None:
+            if criteria.startswith('//'):
+                return len(driver.find_elements_by_xpath(criteria)) > 0
+            return len(driver.find_elements_by_accessibility_id(criteria)) > 0
+        if prefix == 'name':
+            return len(driver.find_elements_by_name(criteria)) > 0
+        if prefix == 'class':
+            return len(driver.find_elements_by_class_name(criteria)) > 0
+        if prefix == 'xpath':
+            return len(driver.find_elements_by_xpath(criteria)) > 0
+        if prefix == 'accessibility_id':
+            return len(driver.find_elements_by_accessibility_id(criteria)) > 0
+        if prefix == 'image':
+            try:
+                return len(driver.find_elements_by_image(criteria)) > 0
+            except InvalidSelectorException:
+                zoomba.fail("Selecting by image is only available when using Appium v1.18.0 or higher")
         zoomba.fail("Element locator with prefix '" + prefix + "' is not supported")
 
     def _parse_locator(self, locator):
@@ -669,4 +785,18 @@ class DesktopLibrary(AppiumLibrary):
 
     def _wait_until_page_contains_element(self, locator, timeout=None, error=None):
         """Internal version to avoid duplicate screenshots"""
-        AppiumCommon.wait_until_page_contains_element(self, locator, timeout, error)
+        if not error:
+            error = "Element '%s' did not appear in <TIMEOUT>" % locator
+        self._wait_until(timeout, error, self._is_element_present, locator)
+
+    # Overrides to prevent expensive log_source call
+    def _wait_until_no_error(self, timeout, wait_func, *args):
+        timeout = utils.timestr_to_secs(timeout) if timeout is not None else self._timeout_in_secs
+        max_time = time() + timeout
+        while True:
+            timeout_error = wait_func(*args)
+            if not timeout_error:
+                return
+            if time() > max_time:
+                raise AssertionError(timeout_error)
+            sleep(0.2)
